from .models import SongModel

STANDBY_DEFAULT_MIN_SCORE = 0.5
STANDBY_FULL_SCORE = 1


def get_standby_score(origin, standby):

    # TODO: move this function to utils module
    def duration_ms_to_duration(ms):
        if not ms:  # ms is empty
            return 0
        parts = ms.split(':')
        assert len(parts) in (2, 3), f'invalid duration format: {ms}'
        if len(parts) == 3:
            h, m, s = parts
        else:
            m, s = parts
            h = 0
        return int(h) * 3600 + int(m) * 60 + int(s)

    def get_duration(s):
        return s.duration if isinstance(s, SongModel) else \
            duration_ms_to_duration(s.duration_ms)

    origin_duration = get_duration(origin)

    score = 0
    if not (origin.album_name and origin_duration):
        score_dis = [4, 4, 1, 1]
    else:
        score_dis = [3, 2, 2, 3]

    if origin.artists_name == standby.artists_name:
        score += score_dis[0]
    if origin.title == standby.title:
        score += score_dis[1]
    # Only compare album_name when it is not empty.
    if origin.album_name and origin.album_name == standby.album_name:
        score += score_dis[2]
    standby_duration = get_duration(standby)
    if origin_duration and \
            abs(origin_duration - standby_duration) / max(origin_duration, 1) < 0.1:
        score += score_dis[3]

    # Debug code for score function
    # print(f"{score}\t('{standby.title}', "
    #       f"'{standby.artists_name}', "
    #       f"'{standby.album_name}', "
    #       f"'{standby.duration_ms}')")
    return score / 10



from PyQt5.QtWidgets import QWidget, QVBoxLayout, QLabel, QHBoxLayout

from feeluown.gui.widgets.textbtn import TextButton
from feeluown.gui.helpers import ClickableMixin


class ClickableHeader(ClickableMixin, QWidget):
    btn_text_fold = '△'
    btn_text_unfold = '▼'

    def __init__(self, header, checked=False, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._is_checked = False

        self.inner_header = header
        self.btn = TextButton(self._get_btn_text(self._is_checked))

        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self.inner_header)
        self._layout.addStretch(0)
        self._layout.addWidget(self.btn)

        self.btn.clicked.connect(self.clicked.emit)
        self.clicked.connect(self.toggle)

    def toggle(self):
        self._is_checked = not self._is_checked
        self.btn.setText(self._get_btn_text(self._is_checked))

    def _get_btn_text(self, checked):
        return self.btn_text_unfold if checked else self.btn_text_fold


class Accordion(QWidget):
    """

    TODO: should be able to customize spacing.
    TODO: API will be changed.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)

    def add_section(self,
                    header: QLabel,
                    content: QWidget,
                    header_spacing: int,
                    section_spacing: int):

        def toggle_content():
            if content.isVisible():
                content.hide()
            else:
                content.show()

        clickable_header = ClickableHeader(header, not content.isVisible())
        clickable_header.clicked.connect(toggle_content)

        self._layout.addWidget(clickable_header)
        self._layout.addSpacing(header_spacing)
        self._layout.addWidget(content)
        self._layout.addSpacing(section_spacing)



import logging
from typing import TYPE_CHECKING

from PyQt5.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout, QScrollArea, QFrame

from feeluown.gui.widgets.my_music import MyMusicView
from feeluown.gui.widgets.header import LargeHeader, MidHeader
from feeluown.gui.widgets.playlists import PlaylistsView


if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


logger = logging.getLogger(__name__)


async def render(req, identifier, **kwargs):
    app: 'GuiApp' = req.ctx['app']
    provider = app.library.get(identifier)
    app.ui.right_panel.set_body(View(app, provider))


class View(QWidget):
    def __init__(self, app: 'GuiApp', provider, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._app = app
        self._provider = provider

        self.title = LargeHeader(self._provider.name, parent=self)
        self.my_music_header = MidHeader('我的音乐', parent=self)
        self.my_music_view = MyMusicView(parent=self)
        self.playlists_header = MidHeader('歌单列表', parent=self)
        self.playlists_view = PlaylistsView(parent=self)

        self.my_music_view.setModel(self._app.mymusic_uimgr.model)
        self.playlists_view.setModel(self._app.pl_uimgr.model)

        self.playlists_view.show_playlist.connect(
            lambda pl: self._app.browser.goto(model=pl))

        self._playlists_scroll = QScrollArea(self)
        self._playlists_scroll.setWidget(self.playlists_view)
        self._playlists_scroll.setWidgetResizable(True)
        self.playlists_view.setFrameShape(QFrame.NoFrame)

        self._layout = QVBoxLayout(self)
        self._body_layout = QHBoxLayout()
        self._l_layout = QVBoxLayout()
        self._r_layout = QVBoxLayout()

        self._layout.addWidget(self.title)
        self._layout.addLayout(self._body_layout)

        self._body_layout.addLayout(self._l_layout)
        self._body_layout.addLayout(self._r_layout)

        self._l_layout.addWidget(self.my_music_header)
        self._l_layout.addWidget(self.my_music_view)
        self._l_layout.addStretch(0)

        self._r_layout.addWidget(self.playlists_header)
        self._r_layout.addWidget(self._playlists_scroll)
        self._r_layout.addStretch(0)



import sys
from typing import TYPE_CHECKING

from PyQt5.QtCore import QSize, Qt
from PyQt5.QtWidgets import (
    QFrame, QLabel, QVBoxLayout, QScrollArea, QMessageBox,
    QFormLayout, QDialog, QLineEdit, QDialogButtonBox,
)

from feeluown.collection import CollectionAlreadyExists, CollectionType
from feeluown.utils.reader import create_reader, Reader
from feeluown.utils.aio import run_fn
from feeluown.gui.components import CollectionListView
from feeluown.gui.widgets import HomeButton, AIButton
from feeluown.gui.widgets.separator import Separator
from .provider_bar import ProviderBar, ListViewContainer as LVC

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


class LeftPanel(QScrollArea):

    def __init__(self, app: 'GuiApp', parent=None):
        super().__init__(parent)
        self._app = app

        self.setWidgetResizable(True)
        self.setFrameShape(QFrame.NoFrame)

        self.p = _LeftPanel(app, self)
        self.setWidget(self.p)

        if sys.platform.lower() != 'darwin':
            self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        # HELP(cosven): size policy is not working
        # self.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Expanding)
        self.setMaximumWidth(280)

    def sizeHint(self):
        size = super().sizeHint()
        width = min(self._app.width() * 22 // 100, 240)
        return QSize(width, size.height())

    async def show_provider_current_user_playlists(self, provider):
        self.p.playlists_con.show()
        self._app.pl_uimgr.clear()

        playlists = await run_fn(provider.current_user_list_playlists)
        reader = await run_fn(provider.current_user_fav_create_playlists_rd)
        reader = create_reader(reader)
        assert isinstance(reader, Reader)
        fav_playlists = reader.readall()
        self._app.pl_uimgr.add(playlists)
        self._app.pl_uimgr.add(fav_playlists, is_fav=True)


class _LeftPanel(QFrame):

    def __init__(self, app: 'GuiApp', parent=None):
        super().__init__(parent)
        self._app = app

        self.home_btn = HomeButton(height=30, parent=self)
        self.ai_btn = AIButton(height=30, padding=0.2, parent=self)
        self.collections_header = QLabel('本地收藏集', self)
        self.collections_header.setToolTip(
            '我们可以在本地建立『收藏集』来收藏自己喜欢的音乐资源\n\n'
            '每个收藏集都以一个独立 .fuo 文件的存在，'
            '将鼠标悬浮在收藏集上，可以查看文件所在路径。\n'
            '新建 fuo 文件，则可以新建收藏集，文件名即是收藏集的名字。\n\n'
            '手动编辑 fuo 文件即可编辑收藏集中的音乐资源，也可以在界面上拖拽来增删歌曲。'
        )
        self.collections_view = CollectionListView(self._app)
        self.collections_con = LVC(self.collections_header, self.collections_view)
        self._top_separator = Separator(self._app)
        self.provider_bar = ProviderBar(self._app)

        # For backward compatibility.
        self.playlists_con = self.provider_bar.playlists_con
        self.my_music_con = self.provider_bar.my_music_con
        self.playlists_header = self.provider_bar.playlists_header

        self._layout = QVBoxLayout(self)

        self._layout.setSpacing(0)
        self._layout.setContentsMargins(16, 10, 16, 0)
        self._layout.addWidget(self.home_btn)
        self._layout.addWidget(self.ai_btn)
        self._layout.addWidget(self.collections_con)
        self._layout.addWidget(self._top_separator)
        self._layout.addWidget(self.provider_bar)
        self._layout.addStretch(0)

        self.collections_view.show_collection.connect(
            lambda coll: self._app.browser.goto(page=f'/colls/{coll.identifier}'))
        self.collections_view.remove_collection.connect(self._remove_coll)
        self.collections_con.create_btn.clicked.connect(
            self.popup_collection_adding_dialog)
        self.collections_view.setFrameShape(QFrame.NoFrame)
        self.setFrameShape(QFrame.NoFrame)
        self.collections_con.create_btn.show()
        self.provider_bar.fold_top_btn.clicked.connect(self._toggle_top_layout)
        if self._app.config.ENABLE_NEW_HOMEPAGE is True:
            self.home_btn.clicked.connect(
                lambda: self._app.browser.goto(page='/homepage'))
        else:
            self.home_btn.clicked.connect(self.show_library)
        if self._app.ai is None:
            self.ai_btn.setDisabled(True)
            self.ai_btn.setToolTip(
                '你需要安装 Python 三方库 openai，并且配置如下配置项，你就可以使用 AI 助手了\n'
                'config.OPENAI_API_KEY = "sk-xxx"\n'
                'config.OPENAI_API_BASEURL = "http://xxx"\n'
                'config.OPENAI_API_MODEL = "model name"\n'
            )
        else:
            self.ai_btn.clicked.connect(self._app.ui.ai_chat_overlay.show)

    def _toggle_top_layout(self, checked):
        widgets = [self._top_separator, self.collections_con, self.home_btn, self.ai_btn]
        if checked:
            self.provider_bar.fold_top_btn.set_direction('down')
            for w in widgets:
                w.hide()
        else:
            self.provider_bar.fold_top_btn.set_direction('up')
            for w in widgets:
                w.show()

    def popup_collection_adding_dialog(self):
        dialog = QDialog(self)
        # Set WA_DeleteOnClose so that the dialog can be deleted (from self.children).
        dialog.setAttribute(Qt.WA_DeleteOnClose)
        layout = QFormLayout(dialog)
        id_edit = QLineEdit(dialog)
        title_edit = QLineEdit(dialog)
        layout.addRow('ID', id_edit)
        layout.addRow('标题', title_edit)
        button_box = QDialogButtonBox(QDialogButtonBox.Cancel | QDialogButtonBox.Save)
        layout.addRow('', button_box)
        button_box.accepted.connect(dialog.accept)
        button_box.rejected.connect(dialog.reject)

        def create_collection_and_reload():
            fname = id_edit.text()
            title = title_edit.text()
            try:
                self._app.coll_mgr.create(fname, title)
            except CollectionAlreadyExists:
                QMessageBox.warning(self, '警告', f"收藏集 '{fname}' 已存在")
            else:
                self._app.coll_mgr.refresh()

        dialog.accepted.connect(create_collection_and_reload)
        dialog.open()

    def show_library(self):
        coll_library = self._app.coll_mgr.get_coll_library()
        self._app.browser.goto(page=f'/colls/{coll_library.identifier}')

    def show_pool(self):
        coll = self._app.coll_mgr.get(CollectionType.sys_pool)
        self._app.browser.goto(page=f'/colls/{coll.identifier}')

    def _remove_coll(self, coll):

        def do():
            self._app.coll_mgr.remove(coll)
            self._app.coll_mgr.refresh()

        box = QMessageBox(QMessageBox.Warning, '提示', f"确认删除收藏集 '{coll.name}' 吗？",
                          QMessageBox.Yes | QMessageBox.No, self)
        box.accepted.connect(do)
        box.open()



import json

from .models import BriefSongModel, ModelState, fmt_artists_names
from feeluown.utils.utils import elfhash


class AnalyzeError(Exception):
    pass


def create_dummy_brief_song(title, artists_name):
    identifier = elfhash(f'{title}-{artists_name}'.encode('utf-8'))
    return BriefSongModel(
        source='dummy',
        identifier=identifier,
        title=title,
        artists_name=artists_name,
        state=ModelState.not_exists,
    )


def analyze_text(text):
    def json_fn(each):
        try:
            return each['title'], fmt_artists_names(each['artists'])
        except KeyError:
            return None

    def line_fn(line):
        parts = line.split('|')
        if len(parts) == 2 and parts[0]:  # title should not be empty
            return (parts[0], parts[1])
        return None

    try:
        data = json.loads(text)
    except json.JSONDecodeError:
        lines = text.strip().split('\n')
        if lines:
            first_line = lines[0].strip()
            if first_line in ('---', '==='):
                parse_each_fn = line_fn
                items = [each.strip() for each in lines[1:] if each.strip()]
            elif first_line == '```json':
                try:
                    items = json.loads(text[7:-3])
                except json.JSONDecodeError:
                    raise AnalyzeError('invalid JSON content inside code block')
                parse_each_fn = json_fn
            else:
                raise AnalyzeError('invalid JSON content')
    else:
        if not isinstance(data, list):
            # should be like [{"title": "xxx", "artists_name": "yyy"}]
            raise AnalyzeError('content has invalid format')
        parse_each_fn = json_fn
        items = data

    err_count = 0
    songs = []
    for each in items:
        result = parse_each_fn(each)
        if result is not None:
            title, artists_name = result
            song = create_dummy_brief_song(title, artists_name)
            songs.append(song)
        else:
            err_count += 1
    return songs, err_count



import asyncio

from feeluown.utils import aio
from feeluown.utils.dispatch import Signal
from .base_player import AbstractPlayer, State


class PlayerPositionDelegate:
    """
    Player notify others through signal when properties are changed. Mpvplayer
    emit signal in non-main thread, and some receivers need to run in main thread.
    From the profile result(#669), we can see that the cross-thread communication
    costs a lot of CPU resources. When the signal emits too frequently, it
    causes obvious performance degration.

    This delegate runs in main thread, so the issue is almost solved. In addition,
    it emits position_changed signals with lower frequency by default.

    .. versionadded:: 3.8.11
        This class is *experimental*.
    """
    def __init__(self, player: AbstractPlayer, interval=300):
        self._position: float = player.position  # seconds
        self._interval = interval  # microseconds
        self._interval_s = interval / 1000  # seconds

        self._player = player
        self._player.seeked.connect(self._update_position_and_emit)
        self._player.state_changed.connect(self.on_state_changed, aioqueue=True)
        self._player.media_changed.connect(self._update_position_and_emit)

        self.changed = Signal()
        self._should_stop = False

    def initialize(self):
        aio.run_afn(self.start)

    async def start(self):
        while not self._should_stop:
            await asyncio.sleep(self._interval_s)
            self._update_position_and_emit()

    def stop(self):
        self._should_stop = True

    def on_state_changed(self, state):
        # When state is changed, update immediately.
        if state == State.playing:
            self._update_position_and_emit()

    def _update_position_and_emit(self, *_):
        self._position = self._player.position
        self.changed.emit(self._position)



from datetime import datetime

from PyQt5.QtWidgets import QAbstractItemView, QFrame, QVBoxLayout, QScrollArea

from feeluown.library import SearchType
from feeluown.utils.aio import run_afn
from feeluown.gui.components.overlay import AppOverlayContainer
from feeluown.gui.page_containers.table import TableContainer, Renderer
from feeluown.gui.widgets.img_card_list import ImgCardListDelegate
from feeluown.gui.widgets.songs import SongsTableView, ColumnsMode
from feeluown.gui.base_renderer import TabBarRendererMixin
from feeluown.gui.helpers import BgTransparentMixin, unify_scroll_area_style, \
    set_widget_bg_transparent
from feeluown.gui.widgets.magicbox import KeySourceIn, KeyType
from feeluown.gui.widgets.header import LargeHeader, MidHeader
from feeluown.gui.widgets.accordion import Accordion
from feeluown.gui.widgets.labels import MessageLabel
from feeluown.utils.reader import create_reader


Tabs = [('歌曲', SearchType.so),
        ('专辑', SearchType.al),
        ('歌手', SearchType.ar),
        ('歌单', SearchType.pl),
        ('视频', SearchType.vi)]


def get_tab_idx(search_type):
    for i, tab in enumerate(Tabs):
        if tab[1] == search_type:
            return i
    raise ValueError("unknown search type")


def create_search_result_view(app, song):
    q = f'{song.title} {song.artists_name}'
    body = SearchResultView(app, transparent_bg=False)
    view = AppOverlayContainer(app, body, parent=app, adhoc=True)

    source_in = app.browser.local_storage.get(KeySourceIn, None)
    run_afn(body.search_and_render, q, SearchType.so, source_in)
    return view


class SearchResultView(QScrollArea):
    """
    Usage:
        view = SearchResultView(app)
        await view.search_and_render(q, search_type, source_in)
    """
    def __init__(self, app, transparent_bg=True, parent=None):
        super().__init__(parent=parent)

        self.setWidgetResizable(True)
        self.setFrameShape(QFrame.NoFrame)
        unify_scroll_area_style(self)
        if transparent_bg:
            set_widget_bg_transparent(self)

        self.body = Body(app)
        self.setWidget(self.body)

    def fillable_bg_height(self):
        """Implement VFillableBg protocol"""
        return self.body.height() - self.body.accordion.height()

    async def search_and_render(self, q, search_type, source_in):
        await self.body.search_and_render(q, search_type, source_in)


class Body(QFrame, BgTransparentMixin):
    def __init__(self, app, **kwargs):
        super().__init__(**kwargs)

        self._app = app

        self.title = LargeHeader()
        self.hint = MessageLabel()
        self.accordion = Accordion()

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(20, 0, 20, 0)
        self._layout.setSpacing(0)
        self._layout.addSpacing(30)
        self._layout.addWidget(self.title)
        self._layout.addSpacing(10)
        self._layout.addWidget(self.hint)
        self._layout.addSpacing(10)
        self._layout.addWidget(self.accordion)
        self._layout.addStretch(0)

    async def search_and_render(self, q, search_type, source_in):
        # pylint: disable=too-many-locals,too-many-statements
        view = self
        app = self._app

        self.title.setText(f'搜索“{q}”')

        tab_index = get_tab_idx(search_type)
        succeed = 0
        start = datetime.now()
        is_first = True  # Is first search result.
        if source_in is not None:
            source_count = len(source_in)
        else:
            source_count = len(app.library.list())
        hint_msgs = [f'正在搜索 {source_count} 个资源提供方...']
        view.hint.show_msg('\n'.join(hint_msgs))
        async for result in app.library.a_search(
                q, type_in=search_type, source_in=source_in, return_err=True):
            if result.err_msg:
                hint_msgs.append(f'搜索 {result.source} 的资源出错：{result.err_msg}')
                view.hint.show_msg('\n'.join(hint_msgs))
                continue

            table_container = TableContainer(app, view.accordion)
            table_container.layout().setContentsMargins(0, 0, 0, 0)

            # HACK: set fixed row for tables.
            # pylint: disable=protected-access
            for table in table_container._tables:
                assert isinstance(table, QAbstractItemView)
                delegate = table.itemDelegate()
                if isinstance(delegate, ImgCardListDelegate):
                    # FIXME: set fixed_row_count in better way.
                    table._fixed_row_count = 2  # type: ignore[attr-defined]
                    delegate.update_settings("card_min_width", 140)
                elif isinstance(table, SongsTableView):
                    table._fixed_row_count = 8
                    table._row_height = table.verticalHeader().defaultSectionSize()

            renderer = SearchResultRenderer(q, tab_index, source_in=source_in)
            await table_container.set_renderer(renderer)
            _, search_type, attrname, show_handler = renderer.tabs[tab_index]
            objects = getattr(result, attrname) or []
            if not objects:  # Result is empty.
                hint_msgs.append(f'搜索 {result.source} 资源，提供方返回空')
                view.hint.show_msg('\n'.join(hint_msgs))
                continue

            succeed += 1
            if search_type is SearchType.so:
                show_handler(  # type: ignore[operator]
                    create_reader(objects), columns_mode=ColumnsMode.playlist)
            else:
                show_handler(create_reader(objects))  # type: ignore[operator]
            source = objects[0].source
            provider = app.library.get(source)
            provider_name = provider.name
            if is_first is False:
                table_container.hide()
            view.accordion.add_section(MidHeader(provider_name), table_container, 6, 12)
            renderer.meta_widget.hide()
            renderer.toolbar.hide()
            is_first = False
        time_cost = (datetime.now() - start).total_seconds()
        hint_msgs.pop(0)
        hint_msgs.insert(0, f'搜索完成，共有 {succeed} 个有效的结果，花费 {time_cost:.2f}s')
        view.hint.show_msg('\n'.join(hint_msgs))


class SearchResultRenderer(Renderer, TabBarRendererMixin):
    def __init__(self, q, tab_index, source_in=None):
        self.q = q
        self.tab_index = tab_index
        self.source_in = source_in

        self.tabs = [
            (*Tabs[0], 'songs', self.show_songs),
            (*Tabs[1], 'albums', self.show_albums),
            (*Tabs[2], 'artists', self.show_artists),
            (*Tabs[3], 'playlists', self.show_playlists),
            (*Tabs[4], 'videos', self.show_videos),
        ]

    async def render(self):
        self.render_tab_bar()

    def render_by_tab_index(self, tab_index):
        search_type = self.tabs[tab_index][1]
        self._app.browser.local_storage[KeyType] = search_type.value
        query = {'q': self.q, 'type': search_type.value}
        source_in = self._app.browser.local_storage.get(KeySourceIn, None)
        if source_in is not None:
            query['source_in'] = source_in
        self._app.browser.goto(page='/search', query=query)



import logging
from typing import TYPE_CHECKING, Optional
from urllib.parse import urlparse

from requests import ConnectTimeout


if TYPE_CHECKING:
    from feeluown.app import App

logger = logging.getLogger(__name__)


class AlertManager:
    """Monitor app exceptions and send some alerts."""
    def __init__(self):
        # Some alerts handling rely on app and some are not.
        self._app: Optional['App'] = None

    def initialize(self, app: 'App'):
        """"""
        self._app = app
        self._app.player.media_loading_failed.connect(
            self.on_media_loading_failed, aioqueue=True)

    def on_exception(self, e):
        if isinstance(e, ConnectTimeout):
            if e.request is not None:
                url = e.request.url
                hostname = urlparse(url).hostname
            else:
                hostname = ''
            msg = f"链接'{hostname}'超时，请检查你的网络或者代理设置"
            self.show_alert(msg)

    def on_media_loading_failed(self, *_):
        assert self._app is not None
        media = self._app.player.current_media
        if media and media.url:
            proxy = f' {media.http_proxy}' if media.http_proxy else '空'
            hostname = urlparse(media.url).hostname
            msg = (f'无法播放来自 {hostname} 的资源（资源的 HTTP 代理为{proxy}）'
                   '（注：播放引擎无法使用系统代理）')
            self.show_alert(msg)

    def show_alert(self, alert):
        logger.warning(alert)
        self._app.show_msg(alert, timeout=2000)



# flake8: noqa
from .library import Library
from .provider import AbstractProvider, ProviderV2, Provider
from .flags import Flags as ProviderFlags
from .model_state import ModelState
from .model_protocol import (  # deprecated
    BriefSongProtocol,
    BriefVideoProtocol,
    BriefArtistProtocol,
    BriefAlbumProtocol,
    BriefUserProtocol,
    SongProtocol,
)
from .models import ModelFlags, BaseModel, ModelType, SearchType, \
    SongModel, BriefSongModel, \
    BriefArtistModel, BriefAlbumModel, \
    BriefCommentModel, CommentModel, \
    BriefUserModel, UserModel, \
    LyricModel, VideoModel, BriefVideoModel, \
    ArtistModel, AlbumModel, PlaylistModel, BriefPlaylistModel, \
    fmt_artists_names, AlbumType, SimpleSearchResult, \
    get_modelcls_by_type, MediaFlags, \
    V2SupportedModelTypes
from .excs import NoUserLoggedIn, ModelNotFound, \
    ProviderAlreadyExists, ResourceNotFound, MediaNotFound
from .provider_protocol import *
from .uri import (
    Resolver, reverse, resolve, ResolverNotFound, ResolveFailed,
    parse_line, NS_TYPE_MAP,
)
from .collection import Collection, CollectionType



from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QPushButton


class TextButton(QPushButton):
    def __init__(self, *args, height=None, **kwargs):
        """
        .. versionadded:: 3.9
            The *height* argument.
        """
        super().__init__(*args, **kwargs)

        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)
        if height:
            self.setFixedHeight(height)



import json
from functools import wraps

from jsonrpc import JSONRPCResponseManager, Dispatcher

from feeluown.fuoexec.fuoexec import fuoexec_get_globals
from feeluown.serializers import serialize, deserialize
from .base import AbstractHandler


class DynamicDispatcher(Dispatcher):
    def __getitem__(self, key):
        try:
            return self.method_map[key]
        except KeyError:
            method = eval(key, fuoexec_get_globals())  # pylint: disable=eval-used
            return method_wrapper(method)


def method_wrapper(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        new_args = ()
        if args:
            new_args = [deserialize('python', arg) for arg in args]
        new_kwargs = {}
        if kwargs:
            new_kwargs = {}
            for k, v in kwargs.items():
                new_kwargs[k] = deserialize('python', v)
        return func(*new_args, **new_kwargs)
    return wrapper


dispatcher = DynamicDispatcher()
dispatcher.add_method(lambda: "pong", name="ping")


def handle(data):
    response = JSONRPCResponseManager.handle(data, dispatcher)
    if 'result' in response.data:
        result = serialize('python', response.result)
        response.result = result
    return response.data


class JsonRPCHandler(AbstractHandler):
    cmds = 'jsonrpc'

    def handle(self, cmd):
        payload = cmd.args[0]
        return json.dumps(handle(payload))



from typing import TYPE_CHECKING, cast

from PyQt5.QtCore import QEvent, QSize
from PyQt5.QtGui import QResizeEvent
from PyQt5.QtWidgets import (
    QHBoxLayout,
    QVBoxLayout,
    QWidget,
    QTabBar,
    QStackedWidget,
    QSplitter,
)

from feeluown.gui.components import PlayerProgressSliderAndLabel
from feeluown.gui.components.btns import MediaButtonsV2
from feeluown.gui.components.nowplaying import (
    NowplayingArtwork,
    NowplayingLyricView,
    NowplayingCommentListView,
    NowplayingPlayerPlaylistView,
    NowplayingSimilarSongsView,
)
from feeluown.gui.components.line_song import TwoLineSongLabel
from feeluown.gui.helpers import set_default_font_families, esc_hide_widget

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


class CtlButtons(QWidget):

    def __init__(self, app: 'GuiApp', parent: 'PlayerPanel'):
        super().__init__(parent=parent)
        self._app = app

        self.media_btns = MediaButtonsV2(app, button_width=36, parent=self)
        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self.media_btns)


class PlayerPanel(QWidget):

    def __init__(self, app: 'GuiApp', parent=None):
        super().__init__(parent=parent)
        self._app = app
        self.artwork_view = NowplayingArtwork(app, self)
        self.title_label = TwoLineSongLabel(app, self)
        self.ctl_btns = CtlButtons(app, parent=self)
        self.progress = PlayerProgressSliderAndLabel(app, parent=self)

        self._layout = QVBoxLayout(self)
        self.setup_ui()

        self.artwork_view.mv_btn.clicked.connect(self.play_mv)
        self.ctl_btns.media_btns.toggle_video_btn.clicked.connect(
            self.keep_and_enter_video_mode
        )
        self._app.player.media_loaded_v2.connect(
            self.on_media_loaded, aioqueue=True
        )
        self._keep_video_mode = False

    def setup_ui(self):
        self._layout.setContentsMargins(0, 0, 11, 0)
        self._layout.setSpacing(0)
        # Put the cover_label in a vboxlayout and add strech around it,
        # so that cover_label's sizehint is respected.
        self._layout.addWidget(self.title_label)
        self._layout.addSpacing(20)
        self._layout.addWidget(self.artwork_view)
        self._layout.setStretch(self._layout.indexOf(self.artwork_view), 1)
        self._layout.addSpacing(20)
        self._layout.addStretch(0)
        self._layout.addWidget(self.progress)
        self._layout.addWidget(self.ctl_btns)

    def play_mv(self):
        self._app.playlist.set_current_model(self._app.playlist.current_song_mv)
        self.enter_video_mode()

    def keep_and_enter_video_mode(self):
        self._keep_video_mode = True
        self.enter_video_mode()

    def unkeep_and_enter_cover_mode(self):
        self._keep_video_mode = False
        self.enter_cover_mode()

    def enter_video_mode(self):
        # FIXME: should call watch_mgr.set_mode
        self._app.watch_mgr.exit_pip_mode()
        self._app.watch_mgr.exit_fullwindow_mode()
        video_widget = self._app.ui.mpv_widget
        video_widget.overlay_auto_visible = True
        if video_widget.parent() == self.artwork_view:
            video_widget.show()
            self.artwork_view.set_body(video_widget)
        else:
            with video_widget.change_parent():
                # Remember always give video_widget a parent first and
                # then show it. Otherwise, mpv may use its mpv-gui to show video,
                # which causes the player crash.
                self.artwork_view.set_body(video_widget)
                video_widget.show()
        self.ctl_btns.hide()
        self.progress.hide()
        video_widget.ctl_bar.clear_adhoc_btns()
        exit_btn = video_widget.ctl_bar.add_adhoc_btn('退出视频模式')
        fullwindow_btn = video_widget.ctl_bar.add_adhoc_btn('窗口全屏')
        exit_btn.clicked.connect(self.unkeep_and_enter_cover_mode)
        fullwindow_btn.clicked.connect(
            lambda: self._app.watch_mgr.
            enter_fullwindow_mode(go_back=self.enter_video_mode)
        )

    def enter_cover_mode(self):
        self.artwork_view.set_body(None)
        self.ctl_btns.show()
        self.progress.show()

    def showEvent(self, a0) -> None:
        # When self is hidden, mpv_widget may be moved to somewhere else.
        # If it is removed, enter cover mode.
        if (
            self._app.ui.mpv_widget.parent() != self.artwork_view
            or self._app.player.video_format is None
        ):
            self.enter_cover_mode()
        else:
            self.enter_video_mode()
        return super().showEvent(a0)

    def sizeHint(self):
        return QSize(500, 400)

    def on_media_loaded(self, properties):
        if bool(properties['video_format']) is True:
            if self._keep_video_mode:
                self.enter_video_mode()
        else:
            self.enter_cover_mode()


class NowplayingOverlay(QWidget):

    def __init__(self, app: 'GuiApp', parent=None):
        super().__init__(parent=parent)
        self._app = app
        self._app.installEventFilter(self)

        self._splitter = QSplitter(self)
        self.player_panel = PlayerPanel(app, parent=self)
        self.stacked_widget = QStackedWidget(parent=self)
        self.tabbar = QTabBar(self)

        self.lyric_view = NowplayingLyricView(app)
        self.comments_view = NowplayingCommentListView(app)
        self.player_playlist_view = NowplayingPlayerPlaylistView(app)
        self.similar_songs_view = NowplayingSimilarSongsView(app)
        self.tabbar.addTab('歌词')
        self.tabbar.addTab('热门评论')
        self.tabbar.addTab('相似歌曲')
        self.tabbar.addTab('播放队列')
        self.stacked_widget.addWidget(self.lyric_view)
        self.stacked_widget.addWidget(self.comments_view)
        self.stacked_widget.addWidget(self.similar_songs_view)
        self.stacked_widget.addWidget(self.player_playlist_view)
        self.stacked_widget.setCurrentIndex(0)

        esc_hide_widget(self)
        self.tabbar.currentChanged.connect(self.stacked_widget.setCurrentIndex)
        self._layout = QHBoxLayout(self)
        self.setup_ui()

    def setup_ui(self):
        set_default_font_families(self.comments_view)
        self.tabbar.setDocumentMode(True)
        self.tabbar.setShape(QTabBar.TriangularEast)
        self.setAutoFillBackground(True)
        self.lyric_view.viewport().setAutoFillBackground(False)
        self.comments_view.viewport().setAutoFillBackground(False)
        self.player_playlist_view.viewport().setAutoFillBackground(False)
        self.similar_songs_view.viewport().setAutoFillBackground(False)

        # Set contents margin explicitly.
        self._splitter.setHandleWidth(0)
        self._splitter.addWidget(self.player_panel)
        self._splitter.addWidget(self.stacked_widget)
        self._layout.setContentsMargins(20, 20, 20, 20)
        self._layout.addWidget(self._splitter)
        self._layout.addWidget(self.tabbar)

    def showEvent(self, e):
        self.resize(self._app.size())
        super().showEvent(e)

    def eventFilter(self, obj, event):
        if self.isVisible() and obj is self._app and event.type() == QEvent.Resize:
            event = cast(QResizeEvent, event)
            self.resize(event.size())
        return False


if __name__ == '__main__':
    from PyQt5.QtWidgets import QWidget

    from feeluown.gui.debug import simple_layout, mock_app
    from feeluown.player import Metadata

    with simple_layout(theme='dark') as layout, mock_app() as app:
        app.playlist.list.return_value = []
        app.size.return_value = QSize(600, 400)
        widget = NowplayingOverlay(app, None)
        widget.resize(600, 400)
        widget.player_panel.title_label.on_metadata_changed(
            Metadata(title='我和我的祖国', artists=['王菲'])
        )
        layout.addWidget(widget)



from .provider_ui import ProviderUiManager
from .uimodels.my_music import MyMusicUiManager
from .uimodels.playlist import PlaylistUiManager

__all__ = (
    'ProviderUiManager',
    'MyMusicUiManager',
    'PlaylistUiManager',
)



import logging
from typing import TYPE_CHECKING

from feeluown.library import BriefSongModel, reverse
from feeluown.library.text2song import analyze_text, AnalyzeError

try:
    from openai import OpenAI, AuthenticationError
except ImportError:
    AI_RADIO_SUPPORTED = False
else:
    AI_RADIO_SUPPORTED = True

if TYPE_CHECKING:
    from feeluown.app import App


logger = logging.getLogger(__name__)


def song2line(song: BriefSongModel):
    line = reverse(song, as_line=True)
    parts = line.split('#', 1)
    if len(parts) >= 2:
        return parts[1]
    return None


class AIRadio:
    def __init__(self, app: 'App'):
        self._app = app

        self._messages = []
        self._messages.append({"role": "system",
                               "content": self._app.config.AI_RADIO_PROMPT})

    def fetch_songs_func(self, _):
        client = OpenAI(
            api_key=self._app.config.OPENAI_API_KEY,
            base_url=self._app.config.OPENAI_API_BASEURL,
        )
        # NOTE(cosven): 经过一些手动测试，我发现“策略”对推荐结果影响很大。
        #
        # 目前的策略是使用多轮对话，让 AI 知道自己之前推荐过什么样的歌曲，
        # 来避免重复推荐。这样的效果是不错的。
        #
        # 尝试过，但效果不好的策略之一：即使把用户不喜欢的歌曲列表提供给 AI，
        # AI 仍然会推不喜欢的歌曲，所以这里不提供用户不喜欢的歌曲列表。所有的模型都这样。
        # 举个例子：kimi 和 deepseek 很喜欢推荐“田馥甄”的“小幸运”，即使我说自己不喜欢，
        # 它还是又很高概率推荐。可能需要一个更强大的 PROMPT :(
        self._messages.append({
            "role": "user",
            "content": (
                f"当前播放列表内容如下：\n{self.get_msg()}")
        })
        try:
            response = client.chat.completions.create(
                model=self._app.config.OPENAI_MODEL,
                messages=self._messages,
            )
        except AuthenticationError:
            logger.exception('AI authentication failed')
            return []
        except Exception:
            logger.exception('AI request failed')
            return []
        msg = response.choices[0].message
        self._messages.append(msg)
        for _msg in self._messages[-2:]:
            logger.info(f"AI radio, message: {dict(_msg)['content']}")
        try:
            songs, err_count = analyze_text(str(msg.content))
        except AnalyzeError:
            logger.exception('Analyze AI response failed')
            return []
        logger.info(f'AI recommend {len(songs)} songs, err_count={err_count}')
        return songs

    def get_msg(self):
        msg_lines = []
        for song in self._app.playlist.list():
            if self._app.playlist.is_bad(song):
                continue
            line = song2line(song)
            if line is not None:
                msg_lines.append(line)
        return '\n'.join(msg_lines)


# For debugging.
if __name__ == '__main__':
    import os
    from unittest.mock import MagicMock

    app = MagicMock()
    app.config.OPENAI_API_KEY = os.environ.get('DEEPSEEK_API_KEY', '')
    app.config.OPENAI_API_BASEURL = 'https://api.deepseek.com'
    app.config.OPENAI_MODEL = 'deepseek-chat'
    app.config.AI_RADIO_PROMPT = '''\
你是一个音乐推荐系统。你根据用户的歌曲列表分析用户的喜好，给用户推荐一些歌。默认推荐5首歌。

有几个注意点
1. 不要推荐与用户播放列表中一模一样的歌曲。不要推荐用户不喜欢的歌曲。不要重复推荐。
2. 你返回的内容只应该有 JSON，其它信息都不需要。也不要用 markdown 格式返回。
3. 你推荐的歌曲需要使用类似这样的 JSON 格式
    [{"title": "xxx", "artists": ["yyy", "zzz"], "description": "推荐理由"}]
'''
    radio = AIRadio(app)
    radio.get_msg = MagicMock(return_value='''
雨蝶 - 李翊君 - 经典再回首 - 03:52
海阔天空 - Beyond - 华纳超极品音色系列 - 03:59
等待 - 韩磊 - 帝王之声 - 03:39
向天再借五百年 - 韩磊 - 向天再借五百年 - 03:12
你是我的眼 - 萧煌奇 - 你是我的眼 - 05:21
''')
    songs = radio.fetch_songs_func(5)
    for song in songs:
        print(song)
